#!/usr/bin/python3
import os
import dbus
import dbus.service
import logging

from gettext import gettext as _
from .backend import InstallBackend
from .Core.loop import mainloop
from .Core.utils import humanize_size
from SystemUpdater.Core.utils import (
    unLockedEnableShutdown,
)

UPDATER_DBUS_INTERFACE = 'com.kylin.systemupgrade.interface'
UPDATER_DBUS_PATH = '/com/kylin/systemupgrade'
UPDATER_DBUS_SERVICE = 'com.kylin.systemupgrade'

#颜色设置
COLORLOG_SUFFIX = "\033[0m"

# RED_PRE    = "\033[31;1m"
# GREEN_PRE  = "\033[32;1m"
# YELLOW_PRE = "\033[33;1m"
# BULE_PRE   = "\033[34;1m"
# RED_PRE    = "\033[35;1m"
# BACK_BLACK_PER  = "\033[40;1m"
# BACK_RED_PER    = "\033[41;1m"
# BACK_GREEN_PER  = "\033[42;1m"
BACK_YELLOW_PER = "\033[43;1m"
# BACK_BLUE_PER   = "\033[44;1m"
# GREEN_GREENBACK_PRE = "\033[1;32;42m"
BULE_GREENBACK_PRE = "\033[1;34;42m"

COLORLOG_PREFIX     = BACK_YELLOW_PER                                                                    
COLORMETHOR_PREFIX  = BULE_GREENBACK_PRE

#dbus 建立
class UpdateManagerDbusController(dbus.service.Object):
    """ this is a helper to provide the UpdateManagerIFace """

    UPDATER_DBUS_INTERFACE = 'com.kylin.systemupgrade.interface'

    def __init__(self, parent, bus_name,
                object_path=UPDATER_DBUS_PATH):
        dbus.service.Object.__init__(self, bus_name, object_path)
        self.parent = parent
        self.bus = dbus.SystemBus()

        self.transaction = None

    #更新important.list的本次升级的列表
    def on_update_important_list(self):
        header = None
        desc = None

        try:
            # logging.info("Start to update Source Template...")
            obj = self.bus.get_object('com.kylin.software.properties', '/com/kylin/software/properties')
            interface = dbus.Interface(obj, dbus_interface='com.kylin.software.properties.interface')
            retval = interface.updateSourceTemplate(timeout=5)
        except Exception as e:
            header = _("Failed to get update software push, please try to update again later")
            desc = str(e)
            logging.error("update sourceTemplate Failed and Error mes:%s"%str(e))
            return False,header,desc
        
        if retval == False:
            header = _("Failed to get update software push, please try to update again later")
            desc = _("Connection exception with push server")
            logging.warning(header + desc)
            return retval,header,desc
        else:
            # logging.info("Update sourceTemplate successed...")
            return retval,header,desc

    #重启aptdeamon后台服务
    def make_aptdeamon_restart(self):
        header = ''
        desc = ''

        try:
            obj = self.bus.get_object('org.debian.apt', '/org/debian/apt')
            interface = dbus.Interface(obj, dbus_interface='org.debian.apt')
            retval = interface.Quit()
        except Exception as e:
            header = _("aptdeamon restart occur Exception")
            logging.error(str(e))
            return False,header,desc

    def check_connectivity(self):
        header = ''
        desc = ''

        try:
            # logging.info("Start to check network connectivity...")
            obj = self.bus.get_object("org.freedesktop.NetworkManager","/org/freedesktop/NetworkManager")
            interface = dbus.Interface(obj, "org.freedesktop.NetworkManager")
            retval = interface.CheckConnectivity(timeout=5)
            # logging.info("Check network connectivity to Complete...")
            #4 表示网络联通
            if retval == 4:
                return True,header,desc
            else:
                header = _("Please check your network connection and retry.")
                return False,header,desc

        except Exception as e:
            header = _("Please check your network connection and retry.")
            logging.error(str(e))
            return False,header,desc

    @dbus.service.method(UPDATER_DBUS_INTERFACE,
                        in_signature="", out_signature="",
                        sender_keyword="caller_name")
    def Quit(self, caller_name):
        """Request a shutdown of the daemon."""
        #如果在下载就请求 取消
        self.CancelDownload()
        try:
            #尝试接触 关闭枷锁
            unLockedEnableShutdown()
        except:
            logging.error("File lock release failure")
        logging.info("Quitting was requested")
        logging.debug("Quitting main loop...")
        mainloop.quit()
        logging.debug("Exit")

    #Remove all downloaded files.
    @dbus.service.method(UPDATER_DBUS_INTERFACE,out_signature='b')
    def Clean(self):
        try:
            #处于更新和升级中的话 不进行更新
            if self.parent.is_working != InstallBackend.ACTION_DEFUALT_STATUS:
                logging.warning('Clean In the process of updating or Upgrading...')
                return False,'In the process of updating or Upgrading...'
            else:
                logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' Clean ...')
                self.parent.start_clean()
                return True
        except Exception:
            return False

    #获取后端现在的状态
    @dbus.service.method(UPDATER_DBUS_INTERFACE,out_signature='i')
    def GetBackendStatus(self):
        try:
            #处于更新和升级中的话 不进行更新
            logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' GetBackendStatus ...')
            return self.parent.is_working
        except Exception:
            return False

    #apt install -f
    @dbus.service.method(UPDATER_DBUS_INTERFACE,out_signature='b')
    def FixBrokenDepends(self):
        try:
            #处于更新和升级中的话 不进行更新
            if self.parent.is_working != InstallBackend.ACTION_DEFUALT_STATUS:
                logging.warning('FixBrokenDepends In the process of updating or Upgrading...')
                return False,'In the process of updating or Upgrading...'
            else:
                logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' FixBrokenDepends ...')
                self.parent.start_fix_broken()
                return True
        except Exception:
            return False

    #dpkg configure -a 
    @dbus.service.method(UPDATER_DBUS_INTERFACE,out_signature='b')
    def FixIncompleteInstall(self):
        try:
            #处于更新和升级中的话 不进行更新
            if self.parent.is_working != InstallBackend.ACTION_DEFUALT_STATUS:
                logging.warning('FixIncompleteInstall In the process of updating or Upgrading...')
                return False,'In the process of updating or Upgrading...'
            else:
                logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' FixIncompleteInstall ...')
                self.parent.start_fix_incomplete()
                return True
        except Exception:
            return False

    #更新的dbus
    @dbus.service.method(UPDATER_DBUS_INTERFACE,out_signature='b')
    def UpdateDetect(self):
        try:
            #处于更新和升级中的话 不进行更新
            if self.parent.is_working != InstallBackend.ACTION_DEFUALT_STATUS:
                logging.warning('UpdateDetect In the process of Updating or Upgrading...')
                return False,'In the process of updating or Upgrading...'
            else:
                logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' UpdateDetect ...')
                self.parent.start_update()
                return True
        except Exception:
            return False
    
    #部分升级
    @dbus.service.method(UPDATER_DBUS_INTERFACE,in_signature='bas',out_signature='bs')
    def DistUpgradePartial(self,_is_install,_partial_upgrade_list):
        try:
            is_install = bool(_is_install)
            if not self.parent.update_list:
                logging.warning('Perform \"UpdateDetect\" first')
                return False,'Perform \"UpdateDetect\" first'
            #处于更新和升级中的话 不进行升级
            if self.parent.is_working != InstallBackend.ACTION_DEFUALT_STATUS:
                logging.warning('DistUpgradePartial In the process of updating or Upgrading...')
                return False,'In the process of updating or Upgrading...'
            else:
                partial_upgrade_list = [str(i) for i in _partial_upgrade_list]
                logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' DistUpgradePartial and is_install:%r, partial_upgrade_list:%s. ',is_install,",".join(partial_upgrade_list))
                local_upgrade_groups = self.parent.update_list.local_upgrade_data.upgrade_groups
                local_single_pkgs = self.parent.update_list.local_upgrade_data.single_pkgs

                new_upgrade_list = list(set(partial_upgrade_list) & set(local_upgrade_groups + local_single_pkgs))

                if new_upgrade_list:
                    self.parent.start_install(InstallBackend.MODE_UPGRADE_PARTIAL,is_install,partial_upgrade_list)
                    return True,'dbus upgrading'
                else:
                    logging.warning('input upgrade list(%s) not in local upgrade_list(%s)',partial_upgrade_list,local_upgrade_groups+local_single_pkgs)
                    return False,'upgrade_list is empty'
        except Exception as e:
            return False,e

    #全部升级
    @dbus.service.method(UPDATER_DBUS_INTERFACE,in_signature='b',out_signature='bs')
    def DistUpgradeAll(self,_is_install):
        try:
            is_install = bool(_is_install)
            if not self.parent.update_list:
                logging.warning('Perform \"UpdateDetect\" first')
                return False,'Perform \"UpdateDetect\" first'
            #处于更新和升级中的话 不进行升级
            if self.parent.is_working != InstallBackend.ACTION_DEFUALT_STATUS:
                logging.warning('DistUpgradeAll In the process of updating or Upgrading...')
                return False,'In the process of updating or Upgrading...'
            else:
                logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' DistUpgradeAll and is_install:%r...',is_install)
                self.parent.start_install(InstallBackend.MODE_UPGRADE_ALL,is_install)
                return True,'success'
        except Exception as e:
            return False,str(e)
            
    # 全盘升级 也就是 apt dist-upgrade 升级的方式
    @dbus.service.method(UPDATER_DBUS_INTERFACE,in_signature='b',out_signature='bs')
    def DistUpgradeSystem(self,_is_install):
        try:
            is_install = bool(_is_install)
            #处于更新和升级中的话 不进行升级
            if self.parent.is_working != InstallBackend.ACTION_DEFUALT_STATUS:
                logging.warning('DistUpgradeSystem In the process of updating or Upgrading...')
                return False,'In the process of updating or Upgrading...'
            else:
                logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' DistUpgradeSystem and is_install:%r...',is_install)
                self.parent.start_install(InstallBackend.MODE_UPGRADE_SYSTEM,is_install)
                return True,'success'
        except Exception as e:
            logging.error(False, str(e))
            return (False, str(e))

    #卸载包
    @dbus.service.method(UPDATER_DBUS_INTERFACE,in_signature='as',out_signature='b')
    def PurgePackages(self,_purge_list):
        try:
            purge_list = [str(pkg) for pkg in _purge_list]

            # 处于更新和升级中的话 不进行升级
            if self.parent.is_working != InstallBackend.ACTION_DEFUALT_STATUS:
                self.PurgePackagesFinished(False,_("Other tasks are being updated and upgraded, please uninstall them later."),'')
                logging.warning('PurgePackages In the process of updating or Upgrading...')
                return False
            else:
                logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' DistPurgePackages and purge list is:%s...',purge_list)
                self.parent.start_purge_pkgs(purge_list)
                return True
        except Exception as e:
            return False,str(e)

    # 取消transaction 
    @dbus.service.method(UPDATER_DBUS_INTERFACE, out_signature='b')
    def CancelDownload(self):
        status = False
        try:
            logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' CancelDownload...')
            
            #空时直接返回
            if self.transaction == None:
                logging.info("transaction is none")
                return False
            
            if self.transaction.cancellable == True:
                self.transaction.cancel()
                status = True
                logging.info("dbus-mothod cancel task Success")
            else:
                status = False
                logging.info("cancel task Failed")
                
        except Exception as e:
            logging.error("cancel task Failed" + str(e))
            return status
        return status


    # 安装本地deb包
    @dbus.service.method(UPDATER_DBUS_INTERFACE,in_signature='ssbb',out_signature='b')
    def InstallDebFile(self,source = "unKnown", path = "", _check_local_dep = False, _auto_satisfy = False):
        try:
            check_local_dep = bool(_check_local_dep)
            auto_satisfy = bool(_auto_satisfy)
            deb_path = str(path)
            logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' InstallDebFile and check_local_dep:%r, auto_satisfy:%r.',\
                                        check_local_dep,auto_satisfy)
            self.parent.start_deb_install(deb_path, _check_local_dep, _auto_satisfy)
            return True
        except Exception as e:
            logging.error(str(e))
            return (False)
    
    # dbus接口：向数据库display表中插入数据
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='ss', sender_keyword='sender')
    def InsertInstallState(self, item, value, sender=None):
        try:
            logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' InsertInstallState, options：%s，value：%s' % (item, value))
            self.parent.sqlite3_server.insert_into_display(item, value)
            logging.info("Database inserted successfully，options：%s，value：%s" % (item, value))
            return True
        except Exception as e:
            logging.error("Database insert failed，options：%s，value：%s" % (item, value))
            return False
    
    # # download certain package and its dependencies
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='asi', out_signature='b')
    def DownloadPackages(self, pkgs=[], try_times=1):
        logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' DownloadPackages: %s',pkgs)
        if try_times > 0:
            pass
        else:
            try_times = 1
        times = try_times
        while times > 0:
            self.parent.start_download(pkgs)
            times -= 1
        return True

    # 是否允许关机前更新
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='ss', out_signature='bs')
    def UnattendedUpgradeValue(self, operation, value="false"):
        if operation.lower() != "get" and operation.lower() != "set":
            return False, 'Please input [\"set\", \"value\"] to set. \nor [\"get\"] to get whether updates are allowed before shutdown.'
        if operation == "set":
            try:
                logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' UnattendedUpgradeValue is going to %s [allow_unattended_upgrades_shutdown] value to %s.'%(operation,value))
                self.parent.sqlite3_server.insert_into_display("allow_unattended_upgrades_shutdown", value.lower())
            except Exception as e:
                logging.error(str(e))
                return False,str(e)
        else:
            logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' UnattendedUpgradeValue is going to %s [allow_unattended_upgrades_shutdown] value.'%(operation))
            try:
                value = self.parent.sqlite3_server.select_from_display("allow_unattended_upgrades_shutdown")
                logging.info("[allow_unattended_upgrades_shutdown] value is %s."%(value))
            except Exception as e:
                logging.error(str(e))
                return False,str(e)
        return True,"success"

    # 设置自动更新时间
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='s', out_signature='bs')
    def SetAutoUpgradePeriod(self, period):
        logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' SetAutoUpgradePeriod will be set value %s.'%(period))
        try:
            self.parent.sqlite3_server.insert_into_display("update_period", period.lower())
        except Exception as e:
            logging.error(str(e))
        return True,"success"
    
    # # dbus接口：改变apt下载速度
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='sb', out_signature='s')
    def SetDownloadspeedMax(self, speed, set):
        logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' SetDownloadspeedMax, speed:%s, set:%r.'%(speed, set))
        self.parent.is_restart_aptdeamon = True
        if set:
            with open("/etc/apt/apt.conf.d/80apt-download", "w+") as f:
                text = f.write("Acquire::http::Dl-Limit" + " \"" + "%s" % str(speed) + "\";\n")
                text = f.write("Acquire::https::Dl-Limit" + " \"" + "%s" % str(speed) + "\";\n")
                return True
        else:
            if os.path.exists("/etc/apt/apt.conf.d/80apt-download"):
                os.remove("/etc/apt/apt.conf.d/80apt-download")
                return False
            else:
                return False

    # # dbus接口：获取apt下载速度
    @dbus.service.method(UPDATER_DBUS_INTERFACE, out_signature='bs')
    def GetDownloadspeedLimitValue(self):
        logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' GetDownloadspeedLimitValue...')
        try:
            f = open("/etc/apt/apt.conf.d/80apt-download", "r")
            value = f.readline().split()
            return True, str(value[1][1:-2])
        except:
            return False, 0

    #更新进度信息 0~100 进度信息 101为非预期的信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='is')
    def UpdateDetectStatusChanged(self,progress,status):
        logging.info(COLORLOG_PREFIX+"emit"+COLORLOG_SUFFIX+" UpdateDetectStatusChanged progress = %d , status = %s",progress,status)
    
    #替代UpdateDetectFinished 分发更新完成信号，为了解决发送空列表时 控制面板出段错误
    def DistUpdateDetectFinished(self, success, upgrade_group,error_string='',error_desc=''):
        if success == False and upgrade_group == []:
            upgrade_group = ['']
        self.UpdateDetectFinished(success,upgrade_group,error_string,error_desc)
    #更新完成的信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='basss')
    def UpdateDetectFinished(self, success, upgrade_group,error_string='',error_desc=''):
        logging.info(COLORLOG_PREFIX + "emit"+ COLORLOG_SUFFIX + " UpdateDetectFinished success = %r , upgrade_group = %a, error_string = %s , error_desc = %s ",\
                    success,upgrade_group,error_string,error_desc)

    #升级的进度信息 0~100 进度信息 101为非预期的信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='asiss')
    def UpdateDloadAndInstStaChanged(self,groups_list,progress,status,current_details):
        logging.info(COLORLOG_PREFIX + "emit" + COLORLOG_SUFFIX +" UpdateDloadAndInstStaChanged upgrade groups_list = %s progress = %d , status = %s ,current_details = %s\033[0m",\
            groups_list,progress,status,current_details)

    #升级完成的信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='basss')
    def UpdateInstallFinished(self, success, upgrade_group,error_string='',error_desc=''):
        logging.info(COLORLOG_PREFIX + "emit" + COLORLOG_SUFFIX +" UpdateInstallFinished success = %r , upgrade_group = %a, error_string = %s , error_desc = %s ",\
                    success,upgrade_group, error_string,error_desc)

    #发送下载包信息 fix bug 字节大小改成u 无符号32位
    @dbus.service.signal(UPDATER_DBUS_INTERFACE, signature='asiiuui')
    def UpdateDownloadInfo(self,upgrade_group,current_items, total_items, currenty_bytes, total_bytes, current_cps):
        logging.info(COLORLOG_PREFIX + "emit" + COLORLOG_SUFFIX +" UpdateDownloadInfo upgrade_group = %a,current_items = %d, total_items = %d, currenty_bytes = %s, total_bytes = %s, current_cps = %s/s",\
                    upgrade_group,\
                    current_items, total_items, \
                    humanize_size(currenty_bytes), humanize_size(total_bytes),\
                    humanize_size(current_cps))

    #查询解决依赖 信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='bbasasasss')
    def UpdateDependResloveStatus(self, resolver_status, remove_status,remove_pkgs,pkg_raw_description,delete_desc,error_string='',error_desc=''):
        logging.info(COLORLOG_PREFIX + "emit" + COLORLOG_SUFFIX +" UpdateDependResloveStatus:resolver_status = %r , resolver_status = %r , remove_pkgs = %a,pkg_raw_description = %a ,delete_desc = %s,error_string = %s , error_desc = %s ",\
            resolver_status,remove_status,remove_pkgs,pkg_raw_description,delete_desc,error_string,error_desc)

    #查询dist-upgrade解决依赖
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='bbasasss')
    def DistupgradeDependResloveStatus(self, resolver_status, remove_status,remove_pkgs,pkg_raw_description,error_string='',error_desc=''):
        logging.info(COLORLOG_PREFIX + "emit" + COLORLOG_SUFFIX +" DistupgradeDependResloveStatus:resolver_status = %r , resolver_status = %r , remove_pkgs = %a,pkg_raw_description = %a ,error_string = %s , error_desc = %s ",\
            resolver_status,remove_status,remove_pkgs,pkg_raw_description,error_string,error_desc)

    # 信号是否可取消
    @dbus.service.signal(UPDATER_DBUS_INTERFACE, signature='b')
    def Cancelable(self, Cancelable):
        pass
        # logging.info(COLORLOG_PREFIX + "emit" + COLORLOG_SUFFIX +" Cancelable: %r",Cancelable)
    
    # 插入数据库完成后发送
    @dbus.service.signal(UPDATER_DBUS_INTERFACE, signature='ss')
    def UpdateSqlitSingle(self, appname, date):
        logging.info(COLORLOG_PREFIX + "emit" + COLORLOG_SUFFIX +" UpdateSqlitSingle: [ %s ]: date: %s .",\
                    appname, date)

    #更新进度信息 0~100 进度信息 101为非预期的信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='bbisss')
    def FixBrokenStatusChanged(self,finished,success,progress,status,error_string='',error_desc=''):
        logging.info(COLORLOG_PREFIX+"emit"+COLORLOG_SUFFIX+" FixBrokenStatusChanged finished = %r , success = %r,progress = %d , status = %s,error_string = %s , error_desc = %s",\
            finished,success,progress,status,error_string,error_desc)

    #卸载完成的信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='bss')
    def PurgePackagesFinished(self, success,error_string='',error_desc=''):
        logging.info(COLORLOG_PREFIX + "emit"+ COLORLOG_SUFFIX + " PurgePackagesFinished success = %r , error_string = %s , error_desc = %s ",\
                    success,error_string,error_desc)

    #卸载进度信息 0~100 进度信息 101为非预期的信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='iss')
    def PurgePkgStatusChanged(self,progress,status,current_details):
        logging.info(COLORLOG_PREFIX + "emit" + COLORLOG_SUFFIX +" PurgePkgStatusChanged progress = %d , status = %s ,current_details = %s",\
            progress,status,current_details)

    #安装deb包完成的信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='bss')
    def InstalldebFinished(self, success,error_string='',error_desc=''):
        logging.info(COLORLOG_PREFIX + "emit"+ COLORLOG_SUFFIX + " InstalldebFinished success = %r , error_string = %s , error_desc = %s ",\
                    success,error_string,error_desc)
        
    #安装进度信息 0~100 进度信息 101为非预期的信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='bss')
    def InstalldebStatusChanged(self,progress,status,current_details):
        logging.info(COLORLOG_PREFIX + "emit" + COLORLOG_SUFFIX +" InstalldebStatusChanged progress = %d , status = %s ,current_details = %s",\
            progress,status,current_details)
        